@@ -15,15 +15,19 @@ from typing import Tuple
 from typing import Union
 from typing import Callable
 from typing import Optional
+from typing import NamedTuple
 from cftool.misc import safe_execute
 from cftool.misc import shallow_copy_dict
+from cftool.array import arr_type
 from cftool.array import save_images
 from cftool.types import tensor_dict_type
 
 from .common import read_image
 from .common import restrict_wh
 from .common import get_suitable_size
+from .common import Padding
 from .common import APIMixin
+from .common import ReadImageResponse
 from ....zoo import DLZoo
 from ....data import predict_tensor_data
 from ....data import TensorInferenceData
@@ -39,6 +43,10 @@ from ....models.cv.diffusion import DDPM
 from ....models.cv.diffusion import ISampler
 from ....models.cv.ae.common import IAutoEncoder
 from ....models.cv.diffusion.utils import get_timesteps
+from ....models.cv.diffusion.utils import CONCAT_KEY
+from ....models.cv.diffusion.utils import CONCAT_TYPE
+from ....models.cv.diffusion.utils import HYBRID_TYPE
+from ....models.cv.diffusion.utils import CROSS_ATTN_KEY
 from ....models.cv.diffusion.samplers.ddim import DDIMMixin
 from ....models.cv.diffusion.samplers.k_samplers import KSamplerMixin
 
@@ -60,6 +68,20 @@ class switch_sampler_context:
             self.api.switch_sampler(self.m_sampler)
 
 
+class SizeInfo(NamedTuple):
+    factor: int
+    opt_size: int
+
+
+class MaskedCond(NamedTuple):
+    image_res: ReadImageResponse
+    mask: np.ndarray
+    remained_image: np.ndarray
+    remained_mask: np.ndarray
+    mask_cond: Tensor
+    remained_image_cond: Tensor
+
+
 class DiffusionAPI(APIMixin):
     m: DDPM
     sampler: ISampler
@@ -95,6 +117,15 @@ class DiffusionAPI(APIMixin):
         else:
             self.first_stage = m.first_stage.core
 
+    @property
+    def size_info(self) -> SizeInfo:
+        opt_size = self.m.img_size
+        if self.first_stage is None:
+            factor = 1
+        else:
+            factor = self.first_stage.img_size // opt_size
+        return SizeInfo(factor, opt_size)
+
     def get_cond(self, cond: Any) -> Tensor:
         if self.cond_model is None:
             msg = "should not call `get_cond` when `cond_model` is not available"
@@ -140,6 +171,7 @@ class DiffusionAPI(APIMixin):
         export_path: Optional[str] = None,
         *,
         seed: Optional[int] = None,
+        use_seed_resize: bool = False,
         # each variation contains (seed, weight)
         variations: Optional[List[Tuple[int, float]]] = None,
         variation_seed: Optional[int] = None,
@@ -152,6 +184,7 @@ class DiffusionAPI(APIMixin):
         original_size: Optional[Tuple[int, int]] = None,
         alpha: Optional[np.ndarray] = None,
         cond: Optional[Any] = None,
+        cond_concat: Optional[Tensor] = None,
         unconditional_cond: Optional[Any] = None,
         num_steps: Optional[int] = None,
         clip_output: bool = True,
@@ -161,7 +194,7 @@ class DiffusionAPI(APIMixin):
         **kwargs: Any,
     ) -> Tensor:
         if cond is not None:
-            if self.cond_type != "concat" and self.cond_model is not None:
+            if self.cond_type != CONCAT_TYPE and self.cond_model is not None:
                 data = TensorInferenceData(cond, batch_size=batch_size)
                 cond = predict_tensor_data(self.cond_model, data)
         if cond is not None and num_samples != len(cond):
@@ -186,19 +219,19 @@ class DiffusionAPI(APIMixin):
         sampled = []
         kw = dict(num_steps=num_steps, verbose=verbose)
         kw.update(shallow_copy_dict(kwargs))
+        factor, opt_size = self.size_info
         if size is None:
-            size = self.m.img_size, self.m.img_size
+            size = opt_size, opt_size
         else:
-            if self.first_stage is None:
-                factor = 1
-            else:
-                factor = self.first_stage.img_size // self.m.img_size
             size = tuple(map(lambda n: round(n / factor), size))  # type: ignore
         uncond_backup = None
+        unconditional_cond_backup = None
         if self.cond_model is not None and unconditional_cond is not None:
-            uncond_backup = self.sampler.unconditional_cond
-            uncond = self.get_cond(unconditional_cond)
-            self.sampler.unconditional_cond = uncond.to(self.device)
+            uncond_backup = getattr(self.sampler, "uncond")
+            unconditional_cond_backup = getattr(self.sampler, "unconditional_cond")
+            uncond = self.get_cond(unconditional_cond).to(self.device)
+            self.sampler.uncond = uncond.clone()
+            self.sampler.unconditional_cond = uncond.clone()
         with eval_context(self.m):
             with self.amp_context:
                 for batch in iterator:
@@ -217,6 +250,34 @@ class DiffusionAPI(APIMixin):
                             variation_seed,
                             variation_strength,
                         )
+                    if use_seed_resize:
+                        z_original_shape = list(i_z.shape[-2:])
+                        z_opt_shape = list(
+                            map(lambda n: round(n / factor), [opt_size, opt_size])
+                        )
+                        if z_original_shape != z_opt_shape:
+                            dx = (z_original_shape[0] - z_opt_shape[0]) // 2
+                            dy = (z_original_shape[1] - z_opt_shape[1]) // 2
+                            x = z_opt_shape[0] if dx >= 0 else z_opt_shape[0] + 2 * dx
+                            y = z_opt_shape[1] if dy >= 0 else z_opt_shape[1] + 2 * dy
+                            dx = max(-dx, 0)
+                            dy = max(-dy, 0)
+                            i_opt_z_shape = (
+                                len(i_cond),
+                                self.m.in_channels,
+                                *z_opt_shape,
+                            )
+                            i_opt_z, _ = self._set_seed_and_variations(
+                                seed,
+                                lambda: torch.randn(i_opt_z_shape, device=self.device),
+                                lambda noise: noise,
+                                variations,
+                                variation_seed,
+                                variation_strength,
+                            )
+                            i_z[..., dx : dx + x, dy : dy + y] = i_opt_z[
+                                ..., dx : dx + x, dy : dy + y
+                            ]
                     if z_ref is not None and z_ref_mask is not None:
                         if z_ref_noise is not None:
                             i_kw["ref"] = repeat(z_ref)
@@ -231,10 +292,22 @@ class DiffusionAPI(APIMixin):
                         for k, v in i_kw.items():
                             if isinstance(v, torch.Tensor) and v.is_floating_point():
                                 i_kw[k] = v.half()
+                    if cond_concat is not None:
+                        if self.cond_type != HYBRID_TYPE:
+                            raise ValueError(
+                                f"condition type should be `{HYBRID_TYPE}` when "
+                                f"`cond_concat` is provided"
+                            )
+                        i_cond = {
+                            CROSS_ATTN_KEY: i_cond,
+                            CONCAT_KEY: cond_concat,
+                        }
                     i_sampled = self.m.decode(i_z, cond=i_cond, **i_kw)
                     sampled.append(i_sampled.cpu().float())
         if uncond_backup is not None:
-            self.sampler.unconditional_cond = uncond_backup
+            self.sampler.uncond = uncond_backup
+        if unconditional_cond_backup is not None:
+            self.sampler.unconditional_cond = unconditional_cond_backup
         concat = torch.cat(sampled, dim=0)
         if clip_output:
             concat = torch.clip(concat, -1.0, 1.0)
@@ -262,6 +335,61 @@ class DiffusionAPI(APIMixin):
         torch.cuda.empty_cache()
         return concat
 
+    @staticmethod
+    def _txt_cond(t: Union[str, List[str]], n: Optional[int]) -> Tuple[List[str], int]:
+        if n is None:
+            n = 1 if isinstance(t, str) else len(t)
+        if isinstance(t, str):
+            t = [t] * n
+        if len(t) != n:
+            raise ValueError(
+                f"`num_samples` ({n}) should be identical with "
+                f"the number of `txt` ({len(t)})"
+            )
+        return t, n
+
+    @staticmethod
+    def _get_size(
+        size: Optional[Tuple[int, int]],
+        anchor: int,
+        max_wh: int,
+    ) -> Optional[Tuple[int, int]]:
+        if size is None:
+            return None
+        new_size = restrict_wh(*size, max_wh)
+        return tuple(map(get_suitable_size, new_size, (anchor, anchor)))  # type: ignore
+
+    def _get_masked_cond(
+        self,
+        image: Union[str, Image.Image],
+        mask: Union[str, Image.Image],
+        max_wh: int,
+        anchor: int,
+        mask_image_fn: Callable[[np.ndarray, np.ndarray], np.ndarray],
+        mask_cond_fn: Callable[[np.ndarray], Tensor],
+    ) -> MaskedCond:
+        # handle mask stuffs
+        image_res = read_image(image, max_wh, anchor=anchor)
+        mask = read_image(mask, max_wh, anchor=anchor, to_mask=True).image
+        bool_mask = mask >= 0.5
+        remained_mask = (~bool_mask).astype(np.float16 if self.use_half else np.float32)
+        remained_image = mask_image_fn(remained_mask, image_res.image)
+        # construct condition tensor
+        remained_cond = self._get_z(remained_image)
+        latent_shape = remained_cond.shape[-2:]
+        mask_cond = mask_cond_fn(bool_mask)
+        mask_cond = mask_cond.to(torch.float16 if self.use_half else torch.float32)
+        mask_cond = mask_cond.to(self.device)
+        mask_cond = F.interpolate(mask_cond, size=latent_shape)
+        return MaskedCond(
+            image_res,
+            mask,
+            remained_image,
+            remained_mask,
+            mask_cond,
+            remained_cond,
+        )
+
     def txt2img(
         self,
         txt: Union[str, List[str]],
@@ -278,21 +406,8 @@ class DiffusionAPI(APIMixin):
         verbose: bool = True,
         **kwargs: Any,
     ) -> Tensor:
-        input_is_str = isinstance(txt, str)
-        if num_samples is None:
-            num_samples = 1 if input_is_str else len(txt)
-        if isinstance(txt, str):
-            txt = [txt] * num_samples
-        if len(txt) != num_samples:
-            raise ValueError(
-                f"`num_samples` ({num_samples}) should be identical with "
-                f"the number of `txt` ({len(txt)})"
-            )
-        if size is None:
-            new_size = None
-        else:
-            new_size = restrict_wh(*size, max_wh)
-            new_size = tuple(map(get_suitable_size, new_size, (anchor, anchor)))  # type: ignore
+        txt, num_samples = self._txt_cond(txt, num_samples)
+        new_size = self._get_size(size, anchor, max_wh)
         return self.sample(
             num_samples,
             export_path,
@@ -307,49 +422,151 @@ class DiffusionAPI(APIMixin):
             **kwargs,
         )
 
+    def txt2img_two_stage(
+        self,
+        txt: Union[str, List[str]],
+        export_path: Optional[str] = None,
+        *,
+        anchor: int = 64,
+        max_wh: int = 512,
+        img2img_anchor: int = 32,
+        img2img_fidelity: float = 0.2,
+        num_samples: Optional[int] = None,
+        size: Optional[Tuple[int, int]] = None,
+        num_steps: Optional[int] = None,
+        clip_output: bool = True,
+        callback: Optional[Callable[[Tensor], Tensor]] = None,
+        batch_size: int = 1,
+        verbose: bool = True,
+        **kwargs: Any,
+    ) -> Tensor:
+        txt, num_samples = self._txt_cond(txt, num_samples)
+        factor, opt_size = self.size_info
+        bypass_two_stage = False
+        if size is not None:
+            ns = self._get_size(size, anchor, max_wh)
+            bypass_two_stage = ns[0] // factor == ns[1] // factor == opt_size  # type: ignore
+        if size is None or bypass_two_stage:
+            return self.txt2img(
+                txt,
+                export_path,
+                anchor=anchor,
+                max_wh=max_wh,
+                num_samples=num_samples,
+                size=size,
+                num_steps=num_steps,
+                clip_output=clip_output,
+                callback=callback,
+                batch_size=batch_size,
+                verbose=verbose,
+                **kwargs,
+            )
+        first_stage = self.txt2img(
+            txt,
+            anchor=anchor,
+            max_wh=max_wh,
+            num_samples=num_samples,
+            size=(opt_size * factor, opt_size * factor),
+            num_steps=num_steps,
+            clip_output=clip_output,
+            callback=callback,
+            batch_size=batch_size,
+            verbose=verbose,
+            **kwargs,
+        )
+        resized = F.interpolate(first_stage, size=size[::-1], mode="bilinear")
+        return self.img2img(
+            resized,
+            export_path,
+            anchor=img2img_anchor,
+            max_wh=max_wh,
+            fidelity=img2img_fidelity,
+            cond=[txt],
+            num_steps=num_steps,
+            clip_output=clip_output,
+            verbose=verbose,
+            **kwargs,
+        )
+
+    def txt2img_inpainting(
+        self,
+        txt: str,
+        image: Union[str, Image.Image],
+        mask: Union[str, Image.Image],
+        export_path: Optional[str] = None,
+        *,
+        anchor: int = 64,
+        max_wh: int = 512,
+        num_steps: Optional[int] = None,
+        clip_output: bool = True,
+        callback: Optional[Callable[[Tensor], Tensor]] = None,
+        verbose: bool = True,
+        **kwargs: Any,
+    ) -> Tensor:
+        txt_list, num_samples = self._txt_cond(txt, 1)
+        res = self._get_masked_cond(
+            image,
+            mask,
+            max_wh,
+            anchor,
+            lambda remained_mask, img: np.where(remained_mask, img, 0.5),
+            lambda arr: torch.from_numpy(arr),
+        )
+        # sampling
+        z = torch.randn_like(res.remained_image_cond)
+        return self.sample(
+            num_samples,
+            export_path,
+            z=z,
+            original_size=res.image_res.original_size,
+            alpha=res.image_res.alpha,
+            cond=txt_list,
+            cond_concat=torch.cat([res.mask_cond, res.remained_image_cond], dim=1),
+            num_steps=num_steps,
+            clip_output=clip_output,
+            callback=callback,
+            verbose=verbose,
+            **kwargs,
+        )
+
     def outpainting(
         self,
         txt: str,
         image: Union[str, Image.Image],
         export_path: Optional[str] = None,
         *,
-        anchor: int = 32,
+        anchor: int = 64,
         max_wh: int = 512,
-        fidelity: float = 0.2,
-        padding_mode: str = "cv2_telea",
         num_steps: Optional[int] = None,
         clip_output: bool = True,
+        callback: Optional[Callable[[Tensor], Tensor]] = None,
         verbose: bool = True,
         **kwargs: Any,
     ) -> Tensor:
-        with switch_sampler_context(self, "plms"):
-            res = read_image(image, max_wh, anchor=anchor, padding_mode=padding_mode)
-            if res.alpha is None:
-                raise ValueError("`image` should contain alpha channel in outpainting")
-            z = z_ref = self._get_z(res.image)
-            z_ref_mask = F.interpolate(
-                torch.from_numpy(res.alpha).to(z_ref),
-                z_ref.shape[-2:],
-                mode="bicubic",
-            )
-            return self._img2img(
-                z,
-                export_path,
-                z_ref=z_ref,
-                z_ref_mask=z_ref_mask,
-                fidelity=fidelity,
-                original_size=res.original_size,
-                alpha=None,
-                cond=[txt],
-                num_steps=num_steps,
-                clip_output=clip_output,
-                verbose=verbose,
-                **kwargs,
-            )
+        if isinstance(image, str):
+            image = Image.open(image)
+        if image.mode != "RGBA":
+            raise ValueError("`image` should be `RGBA` in outpainting")
+        *rgb, alpha = image.split()
+        mask = Image.fromarray(255 - np.array(alpha))
+        image = Image.merge("RGB", rgb)
+        return self.txt2img_inpainting(
+            txt,
+            image,
+            mask,
+            export_path,
+            anchor=anchor,
+            max_wh=max_wh,
+            num_steps=num_steps,
+            clip_output=clip_output,
+            callback=callback,
+            verbose=verbose,
+            **kwargs,
+        )
 
     def img2img(
         self,
-        image: Union[str, Image.Image],
+        image: Union[str, Image.Image, Tensor],
         export_path: Optional[str] = None,
         *,
         anchor: int = 32,
@@ -362,14 +579,21 @@ class DiffusionAPI(APIMixin):
         verbose: bool = True,
         **kwargs: Any,
     ) -> Tensor:
-        res = read_image(image, max_wh, anchor=anchor)
-        z = self._get_z(res.image)
+        if isinstance(image, Tensor):
+            original_size = tuple(image.shape[-2:][::-1])
+        else:
+            res = read_image(image, max_wh, anchor=anchor)
+            image = res.image
+            original_size = res.original_size
+            if alpha is None:
+                alpha = res.alpha
+        z = self._get_z(image)
         return self._img2img(
             z,
             export_path,
             fidelity=fidelity,
-            original_size=res.original_size,
-            alpha=res.alpha if alpha is None else alpha,
+            original_size=original_size,  # type: ignore
+            alpha=alpha,
             cond=cond,
             num_steps=num_steps,
             clip_output=clip_output,
@@ -394,33 +618,28 @@ class DiffusionAPI(APIMixin):
     ) -> Tensor:
         # inpainting callback, will not trigger in refine stage
         def callback(out: Tensor) -> Tensor:
-            final = torch.from_numpy(remained_image.copy())
-            final += 0.5 * (1.0 + out) * (1.0 - remained_mask)
+            final = torch.from_numpy(res.remained_image.copy())
+            final += 0.5 * (1.0 + out) * (1.0 - res.remained_mask)
             return 2.0 * final - 1.0
 
-        # handle mask stuffs
-        image_res = read_image(image, max_wh, anchor=anchor)
-        mask = read_image(mask, max_wh, anchor=anchor, to_mask=True).image
-        bool_mask = mask >= 0.5
-        remained_mask = (~bool_mask).astype(np.float16 if self.use_half else np.float32)
-        remained_image = remained_mask * image_res.image
-        # construct condition tensor
-        remained_cond = self._get_z(remained_image)
-        latent_shape = remained_cond.shape[-2:]
-        mask_cond = torch.where(torch.from_numpy(bool_mask), 1.0, -1.0)
-        mask_cond = mask_cond.to(torch.float16 if self.use_half else torch.float32)
-        mask_cond = mask_cond.to(self.device)
-        mask_cond = F.interpolate(mask_cond, size=latent_shape)
-        cond = torch.cat([remained_cond, mask_cond], dim=1)
+        res = self._get_masked_cond(
+            image,
+            mask,
+            max_wh,
+            anchor,
+            lambda remained_mask, img: remained_mask * img,
+            lambda arr: torch.where(torch.from_numpy(arr), 1.0, -1.0),
+        )
+        cond = torch.cat([res.remained_image_cond, res.mask_cond], dim=1)
         # refine with img2img
         if refine_fidelity is not None:
-            z = self._get_z(image_res.image)
+            z = self._get_z(res.image_res.image)
             return self._img2img(
                 z,
                 export_path,
                 fidelity=refine_fidelity,
-                original_size=image_res.original_size,
-                alpha=image_res.alpha if alpha is None else alpha,
+                original_size=res.image_res.original_size,
+                alpha=res.image_res.alpha if alpha is None else alpha,
                 cond=cond,
                 num_steps=num_steps,
                 clip_output=clip_output,
@@ -428,13 +647,13 @@ class DiffusionAPI(APIMixin):
                 **kwargs,
             )
         # sampling
-        z = torch.randn_like(remained_cond)
+        z = torch.randn_like(res.remained_image_cond)
         return self.sample(
             1,
             export_path,
             z=z,
-            original_size=image_res.original_size,
-            alpha=image_res.alpha if alpha is None else alpha,
+            original_size=res.image_res.original_size,
+            alpha=res.image_res.alpha if alpha is None else alpha,
             cond=cond,
             num_steps=num_steps,
             clip_output=clip_output,
@@ -538,6 +757,18 @@ class DiffusionAPI(APIMixin):
     ) -> "DiffusionAPI":
         return cls.from_pipeline(ldm_sd(), device, use_amp=use_amp, use_half=use_half)
 
+    @classmethod
+    def from_sd_version(
+        cls,
+        version: str,
+        device: Optional[str] = None,
+        *,
+        use_amp: bool = False,
+        use_half: bool = False,
+    ) -> "DiffusionAPI":
+        m = ldm_sd_version(version)
+        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)
+
     @classmethod
     def from_sd_anime(
         cls,
@@ -549,6 +780,17 @@ class DiffusionAPI(APIMixin):
         m = ldm_sd_anime()
         return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)
 
+    @classmethod
+    def from_sd_inpainting(
+        cls,
+        device: Optional[str] = None,
+        *,
+        use_amp: bool = False,
+        use_half: bool = False,
+    ) -> "DiffusionAPI":
+        m = ldm_sd_inpainting()
+        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)
+
     @classmethod
     def from_celeba_hq(
         cls,
@@ -592,9 +834,9 @@ class DiffusionAPI(APIMixin):
         m = ldm_semantic()
         return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)
 
-    def _get_z(self, img: np.ndarray) -> Tensor:
+    def _get_z(self, img: arr_type) -> Tensor:
         img = 2.0 * img - 1.0
-        z = torch.from_numpy(img)
+        z = img if isinstance(img, Tensor) else torch.from_numpy(img)
         if self.use_half:
             z = z.half()
         z = z.to(self.device)
@@ -750,10 +992,18 @@ def ldm_sd(pretrained: bool = True, **kwargs: Any) -> DLPipeline:
     return _ldm("diffusion/ldm.sd", 64, 4, 4, pretrained=pretrained, **kwargs)
 
 
+def ldm_sd_version(version: str, pretrained: bool = True) -> DLPipeline:
+    return ldm_sd(pretrained, download_name=f"ldm_sd_{version}")
+
+
 def ldm_sd_anime(pretrained: bool = True) -> DLPipeline:
     return ldm_sd(pretrained, download_name="ldm_sd_anime_nai")
 
 
+def ldm_sd_inpainting(pretrained: bool = True, **kw: Any) -> DLPipeline:
+    return _ldm("diffusion/ldm.sd_inpainting", 64, 9, 4, pretrained=pretrained, **kw)
+
+
 def ldm_celeba_hq(pretrained: bool = True) -> DLPipeline:
     return ldm_vq(
         pretrained=pretrained,
@@ -778,7 +1028,7 @@ def ldm_inpainting(pretrained: bool = True) -> DLPipeline:
             num_heads=8,
             num_head_channels=None,
             resample_with_resblock=True,
-            condition_type="concat",
+            condition_type=CONCAT_TYPE,
             first_stage_config=dict(
                 pretrained=False,
                 model_config=dict(
@@ -799,7 +1049,7 @@ def ldm_sr(pretrained: bool = True) -> DLPipeline:
             start_channels=160,
             attention_downsample_rates=[8, 16],
             channel_multipliers=[1, 2, 2, 4],
-            condition_type="concat",
+            condition_type=CONCAT_TYPE,
             first_stage_config=dict(
                 pretrained=False,
             ),
@@ -820,7 +1070,7 @@ def ldm_semantic(pretrained: bool = True) -> DLPipeline:
             num_head_channels=None,
             attention_downsample_rates=[8, 16, 32],
             channel_multipliers=[1, 4, 8],
-            condition_type="concat",
+            condition_type=CONCAT_TYPE,
             condition_model="rescaler",
             condition_config=dict(
                 num_stages=2,
